==> ver.B.b11.02.tar.gz: ver.B.b11.02/env/src/pdb/matrix/mx_cache.c <==
/**********************************************************************
 
	Copyright (C) 2005- Hirohisa MORI <joshua@nichibun.ac.jp>
 
	This program is free software; you can redistribute it 
	and/or modify it under the terms of the GLOBALBASE 
	Library General Public License (G-LGPL) as published by 

	http://www.globalbase.org/
 
	This program is distributed in the hope that it will be 
	useful, but WITHOUT ANY WARRANTY; without even the 
	implied warranty of MERCHANTABILITY or FITNESS FOR A 
	PARTICULAR PURPOSE.

**********************************************************************/

#include	"memory_debug.h"
#include	"matrix.h"

int _debug_flag;

int
mx_cache_boundary(MX_CACHE * c,int * ix);

//#define MXC_DEBUG	1

void *
xx_mxc_alloc(MX_CACHE * c,int size,char * __f,int __l)
{
int blk,r;
MBLK * ret;
	blk = size >> c->mblk_logsize;
	r = size & ((1<<c->mblk_logsize)-1);
	if ( r )
		blk ++;
#ifndef MXC_DEBUG
	if ( _debug_flag || (blk > c->mblk_hashsize-1) ) {
		ret = xx_d_alloc(sizeof(MBLK)+size,__f,__l);
#else
		ret = xx_d_alloc(sizeof(MBLK)+size,__f,__l);
/*
		ret = xx_d_alloc(sizeof(MBLK)+(blk<<c->mblk_logsize),
					__f,__l);
*/
#endif
		ret->size = MBLK_ALLOC;
		ret->alloc_file = __f;
		ret->alloc_line = __l;
		ret->free_file = 0;
		ret->free_line = 0;
		return ret+1;
#ifndef MXC_DEBUG
	}
#endif
	ret = c->mblk_hash[blk];
	if ( ret == 0 ) {
		ret = xx_d_alloc(sizeof(MBLK)+(blk<<c->mblk_logsize),
					__f,__l);
		ret->size = blk;
		ret->cnt = 1;
		ret->alloc_file = __f;
		ret->alloc_line = __l;
		ret->free_file = 0;
		ret->free_line = 0;
		return ret+1;
	}
	c->mblk_hash[blk] = ret->next;
	ret->cnt ++;
	if ( ret->cnt == MBLK_ALLOC )
		ret->cnt --;
	ret->free_file = 0;
	ret->free_line = 0;
	return ret+1;
}

void
xx_mxc_free(MX_CACHE * c,void * ptr,char * __f,int __l)
{
MBLK * p;
int blk;
	p = ptr;
	p --;
	if ( p->free_file )
		er_panic("xx_mxc_free");
	if ( p->size == MBLK_ALLOC ) {
		d_f_ree(p);
		return;
	}
	p->free_file = __f;
	p->free_line = __l;
	blk = p->size;
	if ( blk > c->mblk_hashsize-1 )
		er_panic("mxc_free");
	if ( blk < 0 )
		er_panic("mxc_free2");
	p->next = c->mblk_hash[blk];
	c->mblk_hash[blk] = p;
}

void
mxc_setup(MX_CACHE * c)
{
	if ( c->mblk_logsize == 0 )
		c->mblk_logsize = MBLK_LOGSIZE;
	if ( c->mblk_hashsize == 0 )
		c->mblk_hashsize = MBLK_HASHSIZE;
	c->mblk_hash = d_alloc(sizeof(MBLK*)*c->mblk_hashsize);
	memset(c->mblk_hash,0,sizeof(MBLK*)*c->mblk_hashsize);
	c->user_header = 0;
}

void
flush_mx_cache(MX_CACHE * c,int f)
{
int i;
MBLK * p;
	if ( c->n == 0 )
		return;
	for ( i = 0 ; i < c->ds_len; i ++ )
		matrix_node_channel_unlock(c->n,c->dirty);
	unlock_node(c->n,0);
	c->n = 0;
	if ( f ) {
		if ( c->user_header ) {
			(*c->user_header->free_func)(c);
			c->user_header = 0;
		}
		for ( i = 0 ; i < c->mblk_hashsize ; i ++ ) {
			for ( ; c->mblk_hash[i] ; ) {
				p = c->mblk_hash[i];
				c->mblk_hash[i] = p->next;
				d_f_ree(p);
			}
		}
	}
}

int
get_mx_cache(MX_CACHE * c,INTEGER64 * src)
{
MATRIX_NODE * n;
int err;
int i;
void * d;
MATRIX_DATA_TYPE * tp;
int ef;
	if ( c->n ) {
		if ( cmp_dim_code(c->m,c->n->dim_code,src) ) {
			flush_mx_cache(c,0);
			goto load;
		}
		c->hit_cnt ++;
	}
	else {
	load:
		c->miss_hit_cnt ++;
		n = 0;
		get_matrix_node_channel
			(&err,&n,c->m,
			src,c->access_ch[0],
			c->gn_tree_node,c->gn_create,0);
		if ( n ) {
			matrix_node_channel_unlock(n,0);
			unlock_node(n,0);
		}
		err = 0;
		n = get_matrix_node_wait(&err,c->m,src,
					c->gn_wait);
		if ( err ) {
			return -1;
		}

//ss_printf("CC %s %s %i\n",pt_dc(c->m,src),pt_dc(c->m,n->dim_code),n->status);
		c->n = n;
		ef = 0;
		if ( c->ds == 0 ) {
			c->ds = d_alloc(sizeof(MATRIX_DH_SET)*
						c->ds_len);
			c->el_sizes = d_alloc(sizeof(int)*c->ds_len);
			ef = 1;
		}
		for ( i = 0 ; i < c->ds_len ; i ++ ) {
			d = get_matrix_node_channel
				(&err,&n,c->m,src,c->access_ch[i],
				GN_NODE,GN_NODE_CREATE,0);
			c->ds[i].tp = tp = c->m->channel_info
					[c->access_ch[i]].data_type;
			if ( c->ds[i].tp->parent ) {
				get_matrix_dh_set(&c->ds[i],d);
			}
			else {
				memset(&c->ds[i],0,sizeof(MATRIX_DH_SET));
				c->ds[i].tp = tp;
				c->ds[i].offset = d;
			}
		}
		if ( ef ) {
			for ( i = 0 ; i < c->ds_len ; i ++ ) {
				tp = c->ds[i].tp;
				if ( tp->parent )
					c->el_sizes[i] = (*tp->parent->get_size)(tp,0);
				else	c->el_sizes[i] = (*tp->get_size)(tp,0);
			}
		}
	}
	return 0;
}

int
mx_cache_boundary(MX_CACHE * c,int * ix)
{
int i,j;
MATRIX_DH_SET * dsp;
	for ( i = 0 ; i < c->ds_len ; i ++ ) {
		dsp = &c->ds[i];
		if ( dsp->ix == 0 )
			return 0;
		if ( dsp->hd == 0 )
			continue;
		for ( j = 0 ; j < c->m->p.dim ; j ++ ) {
			if ( dsp->ix[j] <= ix[j] )
				return -1;
		}
	}
	return 0;
}


int
write_mx_cache(MX_CACHE_PARAM * p)
{
int i;
int * ofs;
//MATRIX_DATA_TYPE * tp;
int el_size;
void * p1;
MATRIX * m;
MX_CACHE * c;
INTEGER64 ix,st;
MATRIX_DH_SET * dsp;
int pp,xx;

	if ( get_mx_cache(p->c,p->dc) < 0 )
		return -1;
	c = p->c;
	m = c->m;
	ofs = mxc_alloc(c,sizeof(int)*m->p.dim);
	if ( p->ofs ) {
		for ( i = 0 ; i < m->p.dim ; i ++ ) {
			st = p->dc[i+1] >> (p->dc[0] * m->dim_divide[i]);
			ix = p->ofs[i] + st;
			st &= -(((INTEGER64)1)<<m->block_size[i]);
			ofs[i] = ix - st;
		}
	}
	else {
		for ( i = 0 ; i < m->p.dim ; i ++ )
			ofs[i] = (p->dc[i+1] >> (p->dc[0]
					* m->dim_divide[i])) &
					((((INTEGER64)1) 
						<< m->block_size[i])-1);
	}

	if ( mx_cache_boundary(p->c,ofs) < 0 ) {
		mxc_free(p->c,ofs);
		return -1;
	}
	for ( i = 0 ; i < c->ds_len ; i ++ ) {
		xx = p->data_ix[i].x;
		if ( xx == MXC_INVALID )
			continue;
		pp = p->data_ix[i].p;
		dsp = &c->ds[i];
		if ( dsp->tp->parent ) {
//			tp = c->ds[i].tp->parent;
//			el_size = (*tp->get_size)(tp,0);
			el_size = c->el_sizes[i];
			p1 = ((char*)dsp->offset)
				+ get_seq_from_ix(ofs,dsp->ix,
					m->p.dim) *
					el_size;
			memcpy(p1,
				((char*)p->data_ptrs[pp]) + el_size*xx,
				el_size);
		}
		else {
//			tp = c->ds[i].tp;
//			el_size = (*tp->get_size)(tp,0);
			el_size = c->el_sizes[i];
			memcpy(dsp->offset,
				((char*)p->data_ptrs[pp]) + xx*el_size,el_size);
		}
		c->dirty = NF_DIRTY;
	}
	mxc_free(c,ofs);
	return 0;
}

int
read_mx_cache(MX_CACHE_PARAM * p)
{
int i;
int * ofs;
//MATRIX_DATA_TYPE * tp;
int el_size;
void * p1;
MATRIX * m;
MX_CACHE * c;
INTEGER64 st,ix;
int pp,xx;
MATRIX_DH_SET * dsp;

	if ( get_mx_cache(p->c,p->dc) < 0 )
		return -1;
	c = p->c;
	m = c->m;
	ofs = mxc_alloc(c,sizeof(int)*m->p.dim);
	if ( p->ofs ) {
		for ( i = 0 ; i < m->p.dim ; i ++ ) {
			st = p->dc[i+1] >> (p->dc[0] * m->dim_divide[i]);
			ix = p->ofs[i] + st;
			st &= -(((INTEGER64)1)<<m->block_size[i]);
			ofs[i] = ix - st;
		}
	}
	else {
		for ( i = 0 ; i < m->p.dim ; i ++ )
			ofs[i] = (p->dc[i+1] >> (p->dc[0]
					* m->dim_divide[i])) &
					((((INTEGER64)1)
						<<m->block_size[i])-1);
	}
	if ( (p->flags & MXCF_OPTIM_1) == 0 )
		if ( mx_cache_boundary(p->c,ofs) < 0 ) {
			mxc_free(p->c,ofs);
			return -1;
		}
	for ( i = 0 ; i < c->ds_len ; i ++ ) {
		xx = p->data_ix[i].x;
		if ( xx == MXC_INVALID )
			continue;
		pp = p->data_ix[i].p;
		dsp = &c->ds[i];
		if ( dsp->tp->parent ) {
//			tp = c->ds[i].tp->parent;
//			el_size = (*tp->get_size)(tp,0);
			el_size = c->el_sizes[i];
			p1 = ((char*)dsp->offset)
				+ get_seq_from_ix(ofs,dsp->ix,
					m->p.dim) *
					el_size;
			memcpy(((char*)p->data_ptrs[pp]) + xx*el_size,
				p1,el_size);
		}
		else {
//			tp = c->ds[i].tp;
//			el_size = (*tp->get_size)(tp,0);
			el_size = c->el_sizes[i];
			memcpy(((char*)p->data_ptrs[pp]) + xx*el_size,
				dsp->offset,el_size);
		}
	}
	mxc_free(c,ofs);
	return 0;
}

==> ver.B.b14.01.tar.gz: ver.B.b14.01/env/src/pdb/matrix/mx_cache.c <==
/**********************************************************************
 
	Copyright (C) 2005- Hirohisa MORI <joshua@nichibun.ac.jp>
 
	This program is free software; you can redistribute it 
	and/or modify it under the terms of the GLOBALBASE 
	Library General Public License (G-LGPL) as published by 

	http://www.globalbase.org/
 
	This program is distributed in the hope that it will be 
	useful, but WITHOUT ANY WARRANTY; without even the 
	implied warranty of MERCHANTABILITY or FITNESS FOR A 
	PARTICULAR PURPOSE.

**********************************************************************/

#include	"memory_debug.h"
#include	"matrix.h"

int _debug_flag;

int
mx_cache_boundary(MX_CACHE * c,int * ix);

//#define MXC_DEBUG	1

void *
xx_mxc_alloc(MX_CACHE * c,int size,char * __f,int __l)
{
int blk,r;
MBLK * ret;
	blk = size >> c->mblk_logsize;
	r = size & ((1<<c->mblk_logsize)-1);
	if ( r )
		blk ++;
#ifndef MXC_DEBUG
	if ( _debug_flag || (blk > c->mblk_hashsize-1) ) {
		ret = xx_d_alloc(sizeof(MBLK)+size,__f,__l);
#else
		ret = xx_d_alloc(sizeof(MBLK)+size,__f,__l);
/*
		ret = xx_d_alloc(sizeof(MBLK)+(blk<<c->mblk_logsize),
					__f,__l);
*/
#endif
		ret->size = MBLK_ALLOC;
		ret->alloc_file = __f;
		ret->alloc_line = __l;
		ret->free_file = 0;
		ret->free_line = 0;
		return ret+1;
#ifndef MXC_DEBUG
	}
#endif
	ret = c->mblk_hash[blk];
	if ( ret == 0 ) {
		ret = xx_d_alloc(sizeof(MBLK)+(blk<<c->mblk_logsize),
					__f,__l);
		ret->size = blk;
		ret->cnt = 1;
		ret->alloc_file = __f;
		ret->alloc_line = __l;
		ret->free_file = 0;
		ret->free_line = 0;
		return ret+1;
	}
	c->mblk_hash[blk] = ret->next;
	ret->cnt ++;
	if ( ret->cnt == MBLK_ALLOC )
		ret->cnt --;
	ret->free_file = 0;
	ret->free_line = 0;
	return ret+1;
}

void
xx_mxc_free(MX_CACHE * c,void * ptr,char * __f,int __l)
{
MBLK * p;
int blk;
	p = ptr;
	p --;
	if ( p->free_file )
		er_panic("xx_mxc_free");
	if ( p->size == MBLK_ALLOC ) {
		d_f_ree(p);
		return;
	}
	p->free_file = __f;
	p->free_line = __l;
	blk = p->size;
	if ( blk > c->mblk_hashsize-1 )
		er_panic("mxc_free");
	if ( blk < 0 )
		er_panic("mxc_free2");
	p->next = c->mblk_hash[blk];
	c->mblk_hash[blk] = p;
}

void
mxc_setup(MX_CACHE * c)
{
	if ( c->mblk_logsize == 0 )
		c->mblk_logsize = MBLK_LOGSIZE;
	if ( c->mblk_hashsize == 0 )
		c->mblk_hashsize = MBLK_HASHSIZE;
	c->mblk_hash = d_alloc(sizeof(MBLK*)*c->mblk_hashsize);
	memset(c->mblk_hash,0,sizeof(MBLK*)*c->mblk_hashsize);
	c->user_header = 0;
}

void
flush_mx_cache(MX_CACHE * c,int f)
{
int i;
MBLK * p;
	if ( c->n == 0 )
		return;
	for ( i = 0 ; i < c->ds_len; i ++ )
		matrix_node_channel_unlock(c->n,c->dirty);
	unlock_node(c->n,0);
	c->n = 0;
	if ( f ) {
		if ( c->user_header ) {
			(*c->user_header->free_func)(c);
			c->user_header = 0;
		}
		for ( i = 0 ; i < c->mblk_hashsize ; i ++ ) {
			for ( ; c->mblk_hash[i] ; ) {
				p = c->mblk_hash[i];
				c->mblk_hash[i] = p->next;
				d_f_ree(p);
			}
		}
	}
}

int
get_mx_cache(MX_CACHE * c,INTEGER64 * src)
{
MATRIX_NODE * n;
int err;
int i;
void * d;
MATRIX_DATA_TYPE * tp;
int ef;
	if ( c->n ) {
		if ( cmp_dim_code(c->m,c->n->dim_code,src) ) {
			flush_mx_cache(c,0);
			goto load;
		}
		c->hit_cnt ++;
	}
	else {
	load:
		c->miss_hit_cnt ++;
		n = 0;
		err = 0;
		get_matrix_node_channel
			(&err,&n,c->m,
			src,c->access_ch[0],
			c->gn_tree_node,c->gn_create,0);
		if ( n && err >= 0 ) {
			matrix_node_channel_unlock(n,0);
			unlock_node(n,0);
		}
		err = 0;
		n = get_matrix_node_wait(&err,c->m,src,
					c->gn_wait);
		if ( err ) {
			return -1;
		}

//ss_printf("CC %s %s %i\n",pt_dc(c->m,src,PTDC_NODE_ID),pt_dc(c->m,n->dim_code,PTDC_NODE_ID),n->status);
		c->n = n;
		ef = 0;
		if ( c->ds == 0 ) {
			c->ds = d_alloc(sizeof(MATRIX_DH_SET)*
						c->ds_len);
			c->el_sizes = d_alloc(sizeof(int)*c->ds_len);
			ef = 1;
		}
		for ( i = 0 ; i < c->ds_len ; i ++ ) {
			d = get_matrix_node_channel
				(&err,&n,c->m,src,c->access_ch[i],
				GN_NODE,GN_NODE_CREATE,0);
			c->ds[i].tp = tp = c->m->channel_info
					[c->access_ch[i]].data_type;
			if ( c->ds[i].tp->parent ) {
				get_matrix_dh_set(&c->ds[i],d);
			}
			else {
				memset(&c->ds[i],0,sizeof(MATRIX_DH_SET));
				c->ds[i].tp = tp;
				c->ds[i].offset = d;
			}
		}
		if ( ef ) {
			for ( i = 0 ; i < c->ds_len ; i ++ ) {
				tp = c->ds[i].tp;
				if ( tp->parent )
					c->el_sizes[i] = (*tp->parent->get_size)(tp,0);
				else	c->el_sizes[i] = (*tp->get_size)(tp,0);
			}
		}
	}
	return 0;
}

int
mx_cache_boundary(MX_CACHE * c,int * ix)
{
int i,j;
MATRIX_DH_SET * dsp;
	for ( i = 0 ; i < c->ds_len ; i ++ ) {
		dsp = &c->ds[i];
		if ( dsp->ix == 0 )
			return 0;
		if ( dsp->hd == 0 )
			continue;
		for ( j = 0 ; j < c->m->p.dim ; j ++ ) {
			if ( dsp->ix[j] <= ix[j] )
				return -1;
		}
	}
	return 0;
}


int
write_mx_cache(MX_CACHE_PARAM * p)
{
int i;
int * ofs;
//MATRIX_DATA_TYPE * tp;
int el_size;
void * p1;
MATRIX * m;
MX_CACHE * c;
INTEGER64 ix,st;
MATRIX_DH_SET * dsp;
int pp,xx;

	if ( get_mx_cache(p->c,p->dc) < 0 )
		return -1;
	c = p->c;
	m = c->m;
	ofs = mxc_alloc(c,sizeof(int)*m->p.dim);
	if ( p->ofs ) {
		for ( i = 0 ; i < m->p.dim ; i ++ ) {
			st = p->dc[i+1] >> (p->dc[0] * m->dim_divide[i]);
			ix = p->ofs[i] + st;
			st &= -(((INTEGER64)1)<<m->block_size[i]);
			ofs[i] = ix - st;
		}
	}
	else {
		for ( i = 0 ; i < m->p.dim ; i ++ )
			ofs[i] = (p->dc[i+1] >> (p->dc[0]
					* m->dim_divide[i])) &
					((((INTEGER64)1) 
						<< m->block_size[i])-1);
	}

	if ( mx_cache_boundary(p->c,ofs) < 0 ) {
		mxc_free(p->c,ofs);
		return -1;
	}
	for ( i = 0 ; i < c->ds_len ; i ++ ) {
		xx = p->data_ix[i].x;
		if ( xx == MXC_INVALID )
			continue;
		pp = p->data_ix[i].p;
		dsp = &c->ds[i];
		if ( dsp->tp->parent ) {
//			tp = c->ds[i].tp->parent;
//			el_size = (*tp->get_size)(tp,0);
			el_size = c->el_sizes[i];
			p1 = ((char*)dsp->offset)
				+ get_seq_from_ix(ofs,dsp->ix,
					m->p.dim) *
					el_size;
			memcpy(p1,
				((char*)p->data_ptrs[pp]) + el_size*xx,
				el_size);
		}
		else {
//			tp = c->ds[i].tp;
//			el_size = (*tp->get_size)(tp,0);
			el_size = c->el_sizes[i];
			memcpy(dsp->offset,
				((char*)p->data_ptrs[pp]) + xx*el_size,el_size);
		}
		c->dirty = NF_DIRTY;
	}
	mxc_free(c,ofs);
	return 0;
}

int
read_mx_cache(MX_CACHE_PARAM * p)
{
int i;
int * ofs;
//MATRIX_DATA_TYPE * tp;
int el_size;
void * p1;
MATRIX * m;
MX_CACHE * c;
INTEGER64 st,ix;
int pp,xx;
MATRIX_DH_SET * dsp;

	if ( get_mx_cache(p->c,p->dc) < 0 )
		return -1;
	c = p->c;
	m = c->m;
	ofs = mxc_alloc(c,sizeof(int)*m->p.dim);
	if ( p->ofs ) {
		for ( i = 0 ; i < m->p.dim ; i ++ ) {
			st = p->dc[i+1] >> (p->dc[0] * m->dim_divide[i]);
			ix = p->ofs[i] + st;
			st &= -(((INTEGER64)1)<<m->block_size[i]);
			ofs[i] = ix - st;
		}
	}
	else {
		for ( i = 0 ; i < m->p.dim ; i ++ )
			ofs[i] = (p->dc[i+1] >> (p->dc[0]
					* m->dim_divide[i])) &
					((((INTEGER64)1)
						<<m->block_size[i])-1);
	}
	if ( (p->flags & MXCF_OPTIM_1) == 0 )
		if ( mx_cache_boundary(p->c,ofs) < 0 ) {
			mxc_free(p->c,ofs);
			return -1;
		}
	for ( i = 0 ; i < c->ds_len ; i ++ ) {
		xx = p->data_ix[i].x;
		if ( xx == MXC_INVALID )
			continue;
		pp = p->data_ix[i].p;
		dsp = &c->ds[i];
		if ( dsp->tp->parent ) {
//			tp = c->ds[i].tp->parent;
//			el_size = (*tp->get_size)(tp,0);
			el_size = c->el_sizes[i];
			p1 = ((char*)dsp->offset)
				+ get_seq_from_ix(ofs,dsp->ix,
					m->p.dim) *
					el_size;
			memcpy(((char*)p->data_ptrs[pp]) + xx*el_size,
				p1,el_size);
		}
		else {
//			tp = c->ds[i].tp;
//			el_size = (*tp->get_size)(tp,0);
			el_size = c->el_sizes[i];
			memcpy(((char*)p->data_ptrs[pp]) + xx*el_size,
				dsp->offset,el_size);
		}
	}
	mxc_free(c,ofs);
	return 0;
}

void
check_memcpy(char * a,char * b,int size)
{
int i;
int ret;
	ret = 0;
	for ( i = size ; i > 0 ; i -- , a++ ,b ++ )
		if ( *a != *b ) {
			ret = 1;
		}
	if ( ret )
		ss_printf("NOT EQUAL\n");
	else	ss_printf("EQU\n");
}



int
write_mx_cache_vector(MX_CACHE_PARAM * _p)
{
int i;
int * ofs;
//MATRIX_DATA_TYPE * tp;
int el_size;
void * p1;
MATRIX * m;
MX_CACHE * c;
INTEGER64 ix,st;
MATRIX_DH_SET * dsp;
int pp,xx;
int len,slen,tlen,plen;
int offset_ix;
char* buffer_ptr;
INTEGER64 * dc;
MX_CACHE_PARAM p;
	p = *_p;

	if ( get_mx_cache(p.c,p.dc) < 0 )
		return -1;
	c = p.c;
	m = c->m;
	ofs = mxc_alloc(c,sizeof(int)*m->p.dim);
	if ( p.ofs ) {
		for ( i = 0 ; i < m->p.dim ; i ++ ) {
			st = p.dc[i+1] >> (p.dc[0] * m->dim_divide[i]);
			ix = p.ofs[i] + st;
			st &= -(((INTEGER64)1)<<m->block_size[i]);
			ofs[i] = ix - st;
		}
	}
	else {
		for ( i = 0 ; i < m->p.dim ; i ++ )
			ofs[i] = (p.dc[i+1] >> (p.dc[0]
					* m->dim_divide[i])) &
					((((INTEGER64)1) 
						<< m->block_size[i])-1);
	}

	if ( mx_cache_boundary(p.c,ofs) < 0 ) {
		mxc_free(p.c,ofs);
		return -1;
	}


//ss_printf("====\n");
	dc = mxc_alloc(c,sizeof(INTEGER64)*(m->p.dim+1));
	for ( i = 0 ; i < m->p.dim ; i ++ ) {
		if ( p.dc[i+1] >= m->pixel_size[i] )
			goto err;
		if ( p.dc[i+1] < 0 )
			goto err;
		dc[i+1] = p.dc[i+1];
	}
	dc[0] = p.dc[0];
	p.dc = dc;
	len = p.vector_len;
	offset_ix = 0;
	if ( len <= 0 )
		goto end;
	for ( ; len > 0 ; ) {

//ss_printf("len %i offset_ix %i\n",len,offset_ix);
		plen = -1;
		for ( i = 0 ; i < c->ds_len ; i ++ ) {
			xx = p.data_ix[i].x;
			if ( xx == MXC_INVALID )
				continue;
			pp = p.data_ix[i].p;
			dsp = &c->ds[i];
			if ( dsp->tp->parent ) {
				tlen = (dsp->ix[0] - ofs[0]) ; //&0xfffffffe;
				if ( tlen < len )
					slen = tlen;
				else	slen = len;
				if ( plen < 0 )
					plen = slen;
				else if ( plen != slen ) {
					goto err;
				}

				el_size = c->el_sizes[i];
				p1 = ((char*)dsp->offset)
					+ get_seq_from_ix(ofs,dsp->ix,
						m->p.dim) *
						el_size;
				buffer_ptr = ((char**)p.data_ptrs[pp])[xx];
				if ( buffer_ptr == 0 )
					continue;
				
/*
ss_printf("BB %x %x %x %x\n",
(buffer_ptr + el_size*offset_ix)[0],
(buffer_ptr + el_size*offset_ix)[1],
(buffer_ptr + el_size*offset_ix)[2],
(buffer_ptr + el_size*offset_ix)[3]);
*/
				memcpy(p1,
					buffer_ptr + el_size*offset_ix,
					el_size*slen);
			}
			else {
				goto err;
			}
			c->dirty = NF_DIRTY;
		}
		if ( plen < -1 ) {
			goto err;
		}
		len -= plen;
		offset_ix += plen;
	
		if ( len <= 0 )
			break;
	
		p.dc[1] += plen << (p.dc[0]*m->dim_divide[0]);
		for ( i = 0 ; i < m->p.dim ; i ++ ) {
			if ( p.dc[i+1] >= m->pixel_size[i] )
				goto err;
			if ( p.dc[i+1] < 0 )
				goto err;
		}
		if ( get_mx_cache(c,p.dc) < 0 )
			goto err;
		ofs[0] = 0;
	}
end:
	mxc_free(c,dc);
	mxc_free(c,ofs);
	return p.vector_len - len;
err:
	mxc_free(c,dc);
	mxc_free(c,ofs);
	return -1;
}



int
read_mx_cache_vector(MX_CACHE_PARAM * _p)
{
int i;
int * ofs;
//MATRIX_DATA_TYPE * tp;
int el_size;
void * p1;
MATRIX * m;
MX_CACHE * c;
INTEGER64 st,ix;
int pp,xx;
MATRIX_DH_SET * dsp;
int len,slen,tlen,plen;
int offset_ix;
char* buffer_ptr;
INTEGER64 * dc;
MX_CACHE_PARAM p;
int err_flag;

	err_flag = 0;
	p = *_p;
	if ( get_mx_cache(p.c,p.dc) < 0 )
		err_flag = 1;
	c = p.c;
	m = c->m;
	ofs = mxc_alloc(c,sizeof(int)*m->p.dim);
	if ( p.ofs ) {
		for ( i = 0 ; i < m->p.dim ; i ++ ) {
			st = p.dc[i+1] >> (p.dc[0] * m->dim_divide[i]);
			ix = p.ofs[i] + st;
			st &= -(((INTEGER64)1)<<m->block_size[i]);
			ofs[i] = ix - st;
		}
	}
	else {
		for ( i = 0 ; i < m->p.dim ; i ++ )
			ofs[i] = (p.dc[i+1] >> (p.dc[0]
					* m->dim_divide[i])) &
					((((INTEGER64)1)
						<<m->block_size[i])-1);
	}
	if ( (p.flags & MXCF_OPTIM_1) == 0 && err_flag == 0 )
		if ( mx_cache_boundary(p.c,ofs) < 0 ) {
			mxc_free(p.c,ofs);
			return -1;
		}

	dc = mxc_alloc(c,sizeof(INTEGER64)*(m->p.dim+1));
	for ( i = 0 ; i < m->p.dim ; i ++ ) {
		if ( p.dc[i+1] >= m->pixel_size[i] )
			goto err;
		if ( p.dc[i+1] < 0 )
			goto err;
		dc[i+1] = p.dc[i+1];
	}
	dc[0] = p.dc[0];
	p.dc = dc;

	len = p.vector_len;
	offset_ix = 0;
	if ( len <= 0 )
		goto end;
	for ( ; len > 0 ; ) {
	
		plen = -1;

		if ( err_flag ) {
			slen = 1<<m->block_size[0];
			if ( slen > len )
				slen = len;
			plen = slen;
		}
		else {
			for ( i = 0 ; i < c->ds_len ; i ++ ) {
				xx = p.data_ix[i].x;
				if ( xx == MXC_INVALID )
					continue;
				pp = p.data_ix[i].p;
				dsp = &c->ds[i];
				if ( dsp->tp->parent ) {
					tlen = (dsp->ix[0] - ofs[0]) ; //&0xfffffffe;
					if ( tlen < len )
						slen = tlen;
					else	slen = len;
					if ( plen < 0 )
						plen = slen;
					else if ( plen != slen ) {
						goto err;
					}

					el_size = c->el_sizes[i];
					p1 = ((char*)dsp->offset)
						+ get_seq_from_ix(ofs,dsp->ix,
							m->p.dim) *
							el_size;
					buffer_ptr = ((char**)p.data_ptrs[pp])[xx];

					memcpy(buffer_ptr + offset_ix*el_size,
						p1,el_size*slen);
				}
				else {
					goto err;
				}
			}
		}
		if ( plen < -1 ) {
			goto err;
		}
		len -= plen;
		offset_ix += plen;
	
		if ( len <= 0 )
			break;
	
		p.dc[1] += plen << (p.dc[0]*m->dim_divide[0]);
		for ( i = 0 ; i < m->p.dim ; i ++ ) {
			if ( p.dc[i+1] >= m->pixel_size[i] )
				goto err;
			if ( p.dc[i+1] < 0 )
				goto err;
		}
		err_flag = 0;
		if ( get_mx_cache(c,p.dc) < 0 )
			err_flag = 1;
		ofs[0] = 0;
	}
end:
	mxc_free(c,dc);
	mxc_free(c,ofs);
	return p.vector_len - len;
err:
	mxc_free(c,dc);
	mxc_free(c,ofs);
	return -1;
}
